#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2015, 2016 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c14526.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Read the Himawari AHI spectral response functions. Data from
http://cimss.ssec.wisc.edu/goes/calibration/SRF/ahi/ 
"""

import logging
LOG = logging.getLogger(__name__)

import os
import numpy as np

from pyspectral.utils import get_central_wave
from pyspectral import get_config


AHI_BAND_NAMES = ['ch1', 'ch2', 'ch3', 'ch4', 'ch5',
                  'ch6', 'ch7', 'ch8', 'ch9', 'ch10',
                  'ch11', 'ch12', 'ch13', 'ch14', 'ch15', 'ch16']

#: Default time format
_DEFAULT_TIME_FORMAT = '%Y-%m-%d %H:%M:%S'

#: Default log format
_DEFAULT_LOG_FORMAT = '[%(levelname)s: %(asctime)s : %(name)s] %(message)s'


class AhiRSR(object):

    """Container for the Himawari AHI relative spectral response data"""

    def __init__(self, bandname, satname='himawari8'):
        """
        """
        self.satname = satname
        self.bandname = bandname
        self.filenames = {}
        self.rsr = None

        conf = get_config()
        options = {}
        for option, value in conf.items(self.satname + '-ahi', raw=True):
            options[option] = value

        for option, value in conf.items('general', raw=True):
            options[option] = value

        self.output_dir = options.get('rsr_dir', './')

        self._get_bandfilenames(**options)
        LOG.debug("Filenames: " + str(self.filenames))
        if os.path.exists(self.filenames[bandname]):
            self.requested_band_filename = self.filenames[bandname]
            self._load()
        else:
            raise IOError("Couldn't find an existing file for this band: " +
                          str(self.bandname))

        # To be compatible with VIIRS....
        self.filename = self.requested_band_filename

    def _get_bandfilenames(self, **options):
        """Get the AHI rsr filenames"""

        path = options["path"]
        for band in AHI_BAND_NAMES:
            LOG.debug("Band = " + str(band))
            self.filenames[band] = os.path.join(path, options[band])
            LOG.debug(self.filenames[band])
            if not os.path.exists(self.filenames[band]):
                LOG.warning("Couldn't find an existing file for this band: " +
                            str(self.filenames[band]))

    def _load(self):
        """Load the Himawari AHI RSR data for the band requested
        """

        data = np.genfromtxt(self.requested_band_filename,
                             unpack=True,
                             names=['wavenumber',
                                    'response'])
        # Data seems to be wavenumbers (in some unit...)
        # Now, provide wavelengths in micro meters
        wavelength = 1.e4 / data['wavenumber']
        response = data['response']

        idx = np.argsort(wavelength)
        wavelength = np.take(wavelength, idx)
        response = np.take(response, idx)
        self.rsr = {}
        self.rsr['det-1'] = {'wavelength': wavelength,
                             'response': response}


def convert2hdf5(platform_id, sat_number):
    """Retrieve original RSR data and convert to internal hdf5 format"""

    import h5py

    satellite_id = platform_id + str(sat_number)

    ahi = AhiRSR('ch1', satellite_id)
    filename = os.path.join(ahi.output_dir,
                            "rsr_ahi_%s%d.h5" % (platform_id, sat_number))

    with h5py.File(filename, "w") as h5f:
        h5f.attrs['description'] = 'Relative Spectral Responses for AHI'
        h5f.attrs['platform'] = platform_id
        h5f.attrs['sat_number'] = sat_number
        h5f.attrs['band_names'] = AHI_BAND_NAMES

        for chname in AHI_BAND_NAMES:
            ahi = AhiRSR(chname, satellite_id)
            grp = h5f.create_group(chname)
            wvl = ahi.rsr[
                'det-1']['wavelength'][~np.isnan(ahi.rsr['det-1']['wavelength'])]
            rsp = ahi.rsr[
                'det-1']['response'][~np.isnan(ahi.rsr['det-1']['wavelength'])]
            grp.attrs['central_wavelength'] = get_central_wave(wvl, rsp)
            arr = ahi.rsr['det-1']['wavelength']
            dset = grp.create_dataset('wavelength', arr.shape, dtype='f')
            dset.attrs['unit'] = 'm'
            dset.attrs['scale'] = 1e-06
            dset[...] = arr
            arr = ahi.rsr['det-1']['response']
            dset = grp.create_dataset('response', arr.shape, dtype='f')
            dset[...] = arr


if __name__ == "__main__":

    import sys
    LOG = logging.getLogger('ahi_rsr')
    handler = logging.StreamHandler(sys.stderr)

    formatter = logging.Formatter(fmt=_DEFAULT_LOG_FORMAT,
                                  datefmt=_DEFAULT_TIME_FORMAT)
    handler.setFormatter(formatter)
    handler.setLevel(logging.DEBUG)
    LOG.setLevel(logging.DEBUG)
    LOG.addHandler(handler)

    #ahi = AhiRSR('ch1', 'himawari8')
    convert2hdf5('himawari', 8)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <adam.dybbroe@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Convert NOAA 15 Spectral responses to new NOAA STAR format
"""

import numpy as np
import pandas as pd

VIS_FILE = "/home/a000680/data/SpectralResponses/avhrr/NOAA15_AVHRR_responses_VIS.txt"
IR_FILE = "/home/a000680/data/SpectralResponses/avhrr/NOAA15_AVHRR_responses_IR.txt"

VIS_FIELDS = ['wvl1', 'resp1', 'wvl2', 'resp2', 'wvl3a', 'resp3a']
IR_FIELDS = ['wvl3b', 'resp3b', 'wvl4', 'resp4', 'wvl5', 'resp5']

OUTPUT_FILES = {'wvl1': 'NOAA_15_A302xx01.txt',
                'wvl2': 'NOAA_15_A302xx02.txt',
                'wvl3a': 'NOAA_15_A302xx3A.txt',
                'wvl3b': 'NOAA_15_A302xx3B.txt',
                'wvl4': 'NOAA_15_A302xx04.txt',
                'wvl5': 'NOAA_15_A302xx05.txt'
                }


def unpack_data(filename, fields):
    data = pd.read_csv(filename,
                       delim_whitespace=True,
                       names=fields,
                       false_values='-')

    for i in range(0, len(fields), 2):
        wvl = data[:][fields[i]].values
        resp = data[:][fields[i + 1]].values

        wvl = np.ma.masked_where(
            wvl == '-', wvl).compressed().astype('float32')
        resp = np.ma.masked_where(
            resp == '-', resp).compressed().astype('float32') / 100.0

        hdline = "Wavelegth (um)  Normalized RSF"
        np.savetxt(OUTPUT_FILES[fields[i]],
                   np.transpose([wvl.transpose(), resp.transpose()]),
                   fmt='%3.6f',
                   delimiter=' ' * 8,
                   header=hdline, newline='\n')

    return

if __name__ == "__main__":

    unpack_data(IR_FILE, IR_FIELDS)
    unpack_data(VIS_FILE, VIS_FIELDS)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014, 2015, 2016 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <adam.dybbroe@smhi.se>
#   Panu Lahtinen <panu.lahtinen@fmi.fi>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Conversion between radiances and brightness temperatures for the IR bands of
various satellite sensors
"""

import numpy as np
from pyspectral.blackbody import blackbody, blackbody_wn
from pyspectral.utils import BANDNAMES

import logging
LOG = logging.getLogger(__name__)

WAVE_LENGTH = 'wavelength'
WAVE_NUMBER = 'wavenumber'

EPSILON = 0.01
TB_MIN = 150.
TB_MAX = 360.

# Meteosat SEVIRI regression parameters according to documentation
# (PDF_EFFECT_RAD_TO_BRIGHTNESS.pdf).
#
# Tb = C2 * νc/{α * log[C1*νc**3 / L + 1]} - β/α
#
# L = C1 * νc**3 / (exp (C2 νc / [αTb + β]) − 1)
#
# C1 = 2 * h * c**2 and C2 = hc/k
#
# Units are cm-1 for the channel/band central wavenumber, K for the beta
# parameter, and the alpha parameter is dimensionless:
#
SEVIRI = {'IR3.9': {'Meteosat-8': [2567.330, 0.9956, 3.410],
                    'Meteosat-9': [2568.832, 0.9954, 3.438],
                    'Meteosat-10': [],
                    },
          'WV6.2': {'Meteosat-8': [1598.103, 0.9962, 2.218],
                    'Meteosat-9': [1600.548, 0.9963, 2.185],
                    },
          'WV7.3': {'Meteosat-8': [1362.081, 0.9991, 0.478],
                    'Meteosat-9': [1360.330, 0.9991, 0.470],
                    },
          'IR8.7': {'Meteosat-8': [1149.069, 0.9996, 0.179],
                    'Meteosat-9': [1148.620, 0.9996, 0.179],
                    },
          'IR9.7': {'Meteosat-8': [1034.343, 0.9999, 0.060],
                    'Meteosat-9': [1035.289, 0.9999, 0.056],
                    },
          'IR10.8': {'Meteosat-8': [930.647, 0.9983, 0.625],
                     'Meteosat-9': [931.700, 0.9983, 0.640],
                     },
          'IR12.0': {'Meteosat-8': [839.660, 0.9988, 0.397],
                     'Meteosat-9': [836.445, 0.9988, 0.408],
                     },
          'IR13.4': {'Meteosat-8': [752.387, 0.9981, 0.578],
                     'Meteosat-9': [751.792, 0.9981, 0.561],
                     },
          }


class RadTbConverter(object):

    """A radiance to brightness temperature calculator

    It can do the conversion either based on direct use of the band relative
    spectral response function, or on officially (by satellite agencies)
    tabulated standard values using non-linear regression methods.
    Methods:
      1: Spectral response function
      2: non-linear approximation using tabulated coefficients 
    """

    def __init__(self, platform_name, instrument, band, method=1,
                 **options):
        """E.g.:
        platform_name = 'Meteosat-9'
        instrument = 'seviri'
        """
        self.platform_name = platform_name
        self.instrument = instrument
        self.rsr = None
        self.bandname = None
        self.bandwavelength = None

        if isinstance(band, str):
            self.bandname = BANDNAMES.get(band, band)
        elif isinstance(band, (int, long, float)):
            self.bandwavelength = band

        if 'detector' in options:
            self.detector = options['detector']
        else:
            self.detector = 'det-1'

        if 'wavespace' in options:
            if options['wavespace'] not in [WAVE_LENGTH, WAVE_NUMBER]:
                raise AttributeError('Wave space not %s or %s!' % (WAVE_LENGTH,
                                                                   WAVE_NUMBER))
            self.wavespace = options['wavespace']
        else:
            self.wavespace = WAVE_LENGTH

        self._wave_unit = ''
        self._wave_si_scale = 1.0

        if 'tb_resolution' in options:
            self.tb_resolution = options['tb_resolution']
        else:
            self.tb_resolution = 0.1
        self.tb_scale = 1. / self.tb_resolution

        if method == 1:
            self.get_rsr()

        if self.bandname is None:
            epsilon = 0.1
            channel_list = [channel for channel in self.rsr if abs(
                self.rsr[channel]['det-1']['central_wavelength'] - self.bandwavelength) < epsilon]
            self.bandname = BANDNAMES.get(channel_list[0], channel_list[0])

    def get_rsr(self):
        """Get all spectral responses for the sensor"""
        from pyspectral.utils import convert2wavenumber
        from pyspectral.rsr_reader import RelativeSpectralResponse

        sensor = RelativeSpectralResponse(self.platform_name,
                                          self.instrument)
        LOG.debug("Wavenumber? %s", str(self.wavespace))
        if self.wavespace == WAVE_NUMBER:
            LOG.debug("Converting to wavenumber...")
            self.rsr, info = convert2wavenumber(sensor.rsr)
        else:
            self.rsr = sensor.rsr
            info = {'unit': sensor.unit, 'si_scale': sensor.si_scale}

        self._wave_unit = info['unit']
        self._wave_si_scale = info['si_scale']

    def _getsatname(self):
        """Get the satellite name used in the rsr-reader, from the platform
        and number
        """
        if self.platform_name.startswith("Meteosat"):
            return self.platform_name
        else:
            raise NotImplementedError('Platform %s not yet supported...' %
                                      str(self.platform_name))

    def tb2radiance(self, tb_, bandname, lut=None):
        """Get the radiance from the brightness temperature (Tb) given the
        band name. 
        """
        from scipy import integrate

        if self.wavespace == WAVE_NUMBER:
            unit = 'W/m^2 sr^-1 (m^-1)^-1'
            scale = 1.0
        else:
            unit = 'W/m^2 sr^-1 m^-1'
            scale = 1.0

        if not bandname and not np.any(lut):
            raise SyntaxError('Either a band name or a lut needs '
                              'to be provided as input to the function call!')

        if lut:
            ntb = (tb_ * self.tb_scale).astype('int16')
            start = int(lut['tb'][0] * self.tb_scale)
            retv = {}
            bounds = 0, lut['radiance'].shape[0] - 1
            index = np.clip(ntb - start, bounds[0], bounds[1])
            retv['radiance'] = lut['radiance'][index]
            if retv['radiance'].ravel().shape[0] == 1:
                retv['radiance'] = retv['radiance'][0]
            retv['unit'] = unit
            retv['scale'] = scale
            return retv

        if self.wavespace == WAVE_LENGTH:
            wv_ = (self.rsr[bandname][self.detector]['wavelength'] *
                   self._wave_si_scale)
            resp = self.rsr[bandname][self.detector]['response']
            planck = blackbody(wv_, tb_) * resp
        elif self.wavespace == WAVE_NUMBER:
            wv_ = (self.rsr[bandname][self.detector]['wavenumber'] *
                   self._wave_si_scale)
            resp = self.rsr[bandname][self.detector]['response']
            planck = blackbody_wn(wv_, tb_) * resp
        else:
            raise NotImplementedError('%s representation of '
                                      'rsr data not supported!' %
                                      str(self.wavespace))

        radiance = integrate.trapz(planck, wv_) / np.trapz(resp, wv_)
        return {'radiance': radiance,
                'unit': unit,
                'scale': scale}

    def make_tb2rad_lut(self, bandname, filepath):
        """Generate a Tb to radiance look-up table"""
        tb_ = np.arange(TB_MIN, TB_MAX, self.tb_resolution)
        retv = self.tb2radiance(tb_, bandname)
        rad = retv['radiance']
        np.savez(filepath, tb=tb_, radiance=rad.compressed())

    def read_tb2rad_lut(self, filepath):
        """Read the Tb to radiance look-up table"""
        retv = np.load(filepath, 'r')
        return retv

    def tb2radiance_simple(self, tb_, bandname):
        """Get the radiance from the Tb using the simple non-linear regression
        method. SI units of course!
        """
        # L = C1 * νc**3 / (exp (C2 νc / [αTb + β]) − 1)
        #
        # C1 = 2 * h * c**2 and C2 = hc/k
        #
        from pyspectral.blackbody import (H_PLANCK, K_BOLTZMANN, C_SPEED)

        c_1 = 2 * H_PLANCK * C_SPEED ** 2
        c_2 = H_PLANCK * C_SPEED / K_BOLTZMANN

        vc_ = SEVIRI[bandname][self.platform_name][0]
        # Multiply by 100 to get SI units!
        vc_ *= 100.0
        alpha = SEVIRI[bandname][self.platform_name][1]
        beta = SEVIRI[bandname][self.platform_name][2]

        radiance = c_1 * vc_ ** 3 / \
            (np.exp(c_2 * vc_ / (alpha * tb_ + beta)) - 1)

        unit = 'W/m^2 sr^-1 (m^-1)^-1'
        scale = 1.0
        #unit = 'mW/m^2 sr^-1 (cm^-1)^-1'
        #scale = 10.0
        return {'radiance': radiance,
                'unit': unit,
                'scale': scale}

    def radiance2tb_simple(self, rad, bandname):
        """Get the Tb from the radiance using the simple non-linear regression
        method. 
        rad: Radiance in units = 'mW/m^2 sr^-1 (cm^-1)^-1'
        """
        #
        # Tb = C2 * νc/{α * log[C1*νc**3 / L + 1]} - β/α
        #
        # C1 = 2 * h * c**2 and C2 = hc/k
        #
        from pyspectral.blackbody import (H_PLANCK, K_BOLTZMANN, C_SPEED)

        c_1 = 2 * H_PLANCK * C_SPEED ** 2
        c_2 = H_PLANCK * C_SPEED / K_BOLTZMANN

        vc_ = SEVIRI[bandname][self.platform_name][0]
        # Multiply by 100 to get SI units!
        vc_ *= 100.0
        alpha = SEVIRI[bandname][self.platform_name][1]
        beta = SEVIRI[bandname][self.platform_name][2]

        tb_ = c_2 * vc_ / \
            (alpha * np.log(c_1 * vc_ ** 3 / rad + 1)) - beta / alpha

        return tb_

    def radiance2tb(self, rad, wavelength, **kwargs):
        """Get the Tb from the radiance using the Planck function, and optionally the
        relative spectral response function
        rad:
            Radiance in SI units
        """

        from pyspectral.blackbody import blackbody_rad2temp as rad2temp

        return rad2temp(wavelength, rad)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014, 2015, 2016 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <adam.dybbroe@smhi.se>
#   Panu Lahtinen <panu.lahtinen@fmi.fi>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Derive the Near-Infrared reflectance of a given band in the solar and
thermal range (usually the 3.7-3.9 micron band) using a thermal atmospheric
window channel (usually around 11-12 microns).
"""

import os
import numpy as np
from pyspectral.solar import (SolarIrradianceSpectrum,
                              TOTAL_IRRADIANCE_SPECTRUM_2000ASTM)
from pyspectral.utils import BANDNAMES
from pyspectral.radiance_tb_conversion import RadTbConverter
from pyspectral import get_config

import logging
LOG = logging.getLogger(__name__)


WAVE_LENGTH = 'wavelength'
WAVE_NUMBER = 'wavenumber'

EPSILON = 0.01
TB_MIN = 150.
TB_MAX = 360.


class Calculator(RadTbConverter):

    """A thermal near-infrared (e.g. 3.7 micron) band reflectance calculator.

    Given the relative spectral response of the NIR band, the solar zenith
    angle, and the brightness temperatures of the NIR and the Thermal bands,
    derive the solar reflectance for the NIR band removing the thermal
    (terrestrial) part.  The in-band solar flux over the NIR band is
    optional. If not provided, it will be calculated here!

    The relfectance calculated is without units and should be between 0 and 1.
    """

    def __init__(self, platform_name, instrument, band,
                 solar_flux=None, **kwargs):
        """Init"""
        super(Calculator, self).__init__(platform_name, instrument,
                                         band, method=1, **kwargs)

        self.bandname = None
        self.bandwavelength = None

        if isinstance(band, str):
            self.bandname = BANDNAMES.get(band, band)
        elif isinstance(band, (int, long, float)):
            self.bandwavelength = band

        if self.bandname is None:
            epsilon = 0.1
            channel_list = [channel for channel in self.rsr if abs(
                self.rsr[channel]['det-1']['central_wavelength'] - self.bandwavelength) < epsilon]
            self.bandname = BANDNAMES.get(channel_list[0], channel_list[0])

        options = {}
        conf = get_config()
        for option, value in conf.items(platform_name + '-' + instrument,
                                        raw=True):
            options[option] = value

        if solar_flux is None:
            self._get_solarflux()
        else:
            self.solar_flux = solar_flux
        self._rad3x = None
        self._rad3x_t11 = None
        self._solar_radiance = None
        self._rad3x_correction = 1.0
        self._r3x = None
        self._e3x = None

        if 'detector' in kwargs:
            self.detector = kwargs['detector']
        else:
            self.detector = 'det-1'

        resp = self.rsr[self.bandname][self.detector]['response']
        wv_ = self.rsr[self.bandname][self.detector][self.wavespace]
        self.rsr_integral = np.trapz(resp, wv_)

        if 'tb2rad_lut_filename' in options:
            self.lutfile = options['tb2rad_lut_filename']
            if not self.lutfile.endswith('.npz'):
                self.lutfile = self.lutfile + '.npz'

            if not os.path.exists(os.path.dirname(self.lutfile)):
                LOG.warning(
                    "Directory %s does not exist! Check config file" % os.path.dirname(self.lutfile))
                self.lutfile = os.path.join(
                    '/tmp', os.path.basename(self.lutfile))

            LOG.info("lut filename: " + str(self.lutfile))
            if not os.path.exists(self.lutfile):
                self.make_tb2rad_lut(self.bandname,
                                     self.lutfile)
                self.lut = self.read_tb2rad_lut(self.lutfile)
                LOG.info("LUT file created")
            else:
                self.lut = self.read_tb2rad_lut(self.lutfile)
                LOG.info("File was there and has been read!")
        else:
            LOG.info("No lut filename available in config file. "
                     "No lut will be used")
            self.lutfile = None
            self.lut = None

    def derive_rad39_corr(self, bt11, bt13, method='rosenfeld'):
        """Derive the 3.9 radiance correction factor to account for the
        attenuation of the emitted 3.9 radiance by CO2
        absorption. Requires the 11 micron window band and the 13.4
        CO2 absorption band, as e.g. available on SEVIRI. Currently
        only supports the Rosenfeld method
        """
        if method != 'rosenfeld':
            raise AttributeError("Only CO2 correction for SEVIRI using "
                                 "the Rosenfeld equation is supported!")

        LOG.debug("Derive the 3.9 micron radiance CO2 correction coefficent")
        self._rad3x_correction = (bt11 - 0.25 * (bt11 - bt13)) ** 4 / bt11 ** 4

    def _get_solarflux(self):
        """Derive the in-band solar flux from rsr over the Near IR band (3.7
        or 3.9 microns)
        """
        solar_spectrum = \
            SolarIrradianceSpectrum(TOTAL_IRRADIANCE_SPECTRUM_2000ASTM,
                                    dlambda=0.0005,
                                    wavespace=self.wavespace)
        self.solar_flux = \
            solar_spectrum.inband_solarflux(self.rsr[self.bandname])

    def emissive_part_3x(self, tb=True):
        """Get the emissive part of the 3.x band"""
        try:
            # Emissive part:
            # self._e3x = self._rad3x_t11 * (1 -
            self._e3x = self._rad3x * (1 -
                                       np.ma.masked_outside(self._r3x, 0, 1,
                                                            copy=False))
        except TypeError:
            LOG.warning(
                "Couldn't derive the emissive part \n" +
                "Please derive the relfectance prior to requesting the emissive part")

        if tb:
            # FIXME! Get the correct central wavelength
            return self.radiance2tb(self._e3x, 3.9e-6)
        else:
            return self._e3x

    def reflectance_from_tbs(self, sun_zenith, tb_near_ir, tb_thermal,
                             tb_ir_co2=None):
        """The relfectance calculated is without units and should be between 0
        and 1.

        Inputs:

          sun_zenith: Sun zenith angle for every pixel - in degrees

          tb_near_ir: The 3.7 (or 3.9 or equivalent) IR Tb's at every pixel
                      (Kelvin)

          tb_thermal: The 10.8 (or 11 or 12 or equivalent) IR Tb's at every
                      pixel (Kelvin)

          tb_ir_co2: The 13.4 micron channel (or similar - co2 absorption band)
                     brightness temperatures at every pixel. If None, no CO2
                     absorption correction will be applied.

        """
        if np.isscalar(tb_near_ir):
            tb_nir = np.array([tb_near_ir, ])
        else:
            tb_nir = np.array(tb_near_ir)

        if np.isscalar(tb_thermal):
            tb_therm = np.array([tb_thermal, ])
        else:
            tb_therm = np.array(tb_thermal)

        if tb_therm.shape != tb_nir.shape:
            raise ValueError('Dimensions do not match! %s and %s' %
                             (str(tb_therm.shape), str(tb_nir.shape)))

        if tb_ir_co2 is None:
            co2corr = False
            tbco2 = None
        else:
            co2corr = True
            if np.isscalar(tb_ir_co2):
                tbco2 = np.array([tb_ir_co2, ])
            else:
                tbco2 = np.array(tb_ir_co2)

        if self.instrument == 'seviri':
            ch3xname = 'IR3.9'
        elif self.instrument == 'modis':
            ch3xname = '20'
        elif self.instrument in ["avhrr", "avhrr3", "avhrr/3"]:
            ch3xname = 'ch3b'
        elif self.instrument == 'ahi':
            ch3xname = 'ch7'
        elif self.instrument == 'viirs':
            ch3xname = 'M12'
        else:
            raise NotImplementedError('Not yet support for this '
                                      'instrument %s' % str(self.instrument))

        if not self.rsr:
            raise NotImplementedError("Reflectance calculations without "
                                      "rsr not yet supported!")
            # retv = self.tb2radiance_simple(tb_therm, ch3xname)
            # print("tb2radiance_simple conversion: " + str(retv))
            # thermal_emiss_one = retv['radiance']
            # retv = self.tb2radiance_simple(tb_nir, ch3xname)
            # print("tb2radiance_simple conversion: " + str(retv))
            # l_nir = retv['radiance']
        else:
            # Assume rsr in in microns!!!
            # FIXME!
            scale = self.rsr_integral * 1e-6
            retv = self.tb2radiance(tb_therm, ch3xname, self.lut)
            # print("tb2radiance conversion: " + str(retv))
            thermal_emiss_one = retv['radiance'] * scale
            retv = self.tb2radiance(tb_nir, ch3xname, self.lut)
            # print("tb2radiance conversion: " + str(retv))
            l_nir = retv['radiance'] * scale

        if thermal_emiss_one.ravel().shape[0] < 10:
            LOG.info('thermal_emiss_one = %s', str(thermal_emiss_one))
        if l_nir.ravel().shape[0] < 10:
            LOG.info('l_nir = %s', str(l_nir))

        sunz = np.ma.masked_outside(sun_zenith, 0.0, 88.0)
        sunzmask = sunz.mask
        sunz = sunz.filled(88.)

        mu0 = np.cos(np.deg2rad(sunz))
        # mu0 = np.where(np.less(mu0, 0.1), 0.1, mu0)
        self._rad3x = l_nir
        self._rad3x_t11 = thermal_emiss_one
        self._solar_radiance = 1. / np.pi * self.solar_flux * mu0

        # CO2 correction to the 3.9 radiance, only if tbs of a co2 band around
        # 13.4 micron is provided:
        if co2corr:
            self.derive_rad39_corr(tb_therm, tbco2)
        else:
            self._rad3x_correction = 1.0

        # mask = thermal_emiss_one > l_nir

        nomin = l_nir - self._rad3x_t11 * self._rad3x_correction
        LOG.debug("Shapes: %s  %s", str(mu0.shape),
                  str(self._rad3x_t11.shape))
        denom = self._solar_radiance - \
            self._rad3x_t11 * self._rad3x_correction

        r3x = nomin / denom
        # r3x = np.ma.masked_array(r3x, mask=mask)
        # r3x = np.ma.masked_where(r3x < 0, r3x)
        # Do some further masking, also with sun-zenith:
        # r3x = np.ma.masked_outside(r3x, 0.0, 10.0)  # * 100.  # Percent!
        # if np.ma.is_masked(tb_nir):
        #    r3x = np.ma.masked_where(tb_nir.mask, r3x).filled(0)
        self._r3x = np.ma.masked_where(sunzmask, r3x)

        # Reflectances should be between 0 and 1, but values above 1 is
        # perfectly possible and okay! (Multiply by 100 to get reflectances
        # in percent)
        return self._r3x



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <adam.dybbroe@smhi.se>
#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Rayleigh correction of shortwave imager bands in the wavelength range 400 to
800 nm

"""

import logging
import os
import h5py
import numpy as np

from pyspectral import get_config
from pyspectral.rsr_reader import RelativeSpectralResponse
from pyspectral.utils import (BANDNAMES, INSTRUMENTS, get_central_wave,
                              get_rayleigh_reflectance, download_luts,
                              RAYLEIGH_LUT_DIRS)

LOG = logging.getLogger(__name__)

ATMOSPHERES = {'subarctic summer': 4, 'subarctic winter': 5,
               'midlatitude summer': 6, 'midlatitude winter': 7,
               'tropical': 8, 'us-standard': 9}


class BandFrequencyOutOfRange(Exception):
    pass


class Rayleigh(object):

    """Container for the Rayleigh scattering correction of satellite imager short
    wave bands

    """

    def __init__(self, platform_name, sensor, **kwargs):
        self.platform_name = platform_name
        self.sensor = sensor
        self.coeff_filename = None

        if 'atmosphere' in kwargs:
            atm_type = kwargs['atmosphere']
        else:
            atm_type = 'subarctic summer'

        if 'rural_aerosol' in kwargs and kwargs['rural_aerosol']:
            rayleigh_dir = RAYLEIGH_LUT_DIRS['rural_aerosol']
        else:
            rayleigh_dir = RAYLEIGH_LUT_DIRS['rayleigh_only']

        if atm_type not in ATMOSPHERES.keys():
            LOG.error("Atmosphere type %s not supported", atm_type)

        LOG.info("Atmosphere chosen: %s", atm_type)

        conf = get_config()

        options = {}
        for option, value in conf.items('general', raw=True):
            options[option] = value

        # Try fix instrument naming
        instr = INSTRUMENTS.get(platform_name, sensor)
        if instr != sensor:
            sensor = instr
            LOG.warning("Inconsistent sensor/satellite input - " +
                        "sensor set to %s", sensor)

        self.sensor = sensor.replace('/', '')

        # Conversion from standard band names to pyspectral band naming.
        # Preferably take from config! FIXME!
        self.sensor_bandnames = {'B01': 'ch1',
                                 'B02': 'ch2',
                                 'B03': 'ch3',
                                 'M03': 'M3',
                                 'M04': 'M4',
                                 'M05': 'M5',
                                 }

        ext = atm_type.replace(' ', '_')
        lutname = "rayleigh_lut_const_azidiff_%s.h5" % ext
        self.coeff_filename = os.path.join(rayleigh_dir, lutname)

        if not os.path.exists(self.coeff_filename):
            LOG.warning("No lut file %s on disk", self.coeff_filename)
            LOG.info("Will download from internet...")
            download_luts()

        if (not os.path.exists(self.coeff_filename) or
                not os.path.isfile(self.coeff_filename)):
            raise IOError('pyspectral file for Rayleigh scattering correction ' +
                          'does not exist! Filename = ' +
                          str(self.coeff_filename))

        LOG.debug('LUT filename: %s', str(self.coeff_filename))

    def get_effective_wavelength(self, bandname):
        """Get the effective wavelength with Rayleigh scattering in mind"""

        try:
            rsr = RelativeSpectralResponse(self.platform_name, self.sensor)
        except IOError:
            LOG.exception(
                "No spectral responses for this platform and sensor: %s %s", self.platform_name, self.sensor)
            if isinstance(bandname, (float, int, long)):
                LOG.warning(
                    "Effective wavelength is set to the requested band wavelength = %f", bandname)
                return bandname
            raise

        if isinstance(bandname, str):
            bandname = self.sensor_bandnames.get(bandname, bandname)
        elif isinstance(bandname, (float, int, long)):
            if bandname < 0.4 or bandname > 0.8:
                raise BandFrequencyOutOfRange(
                    'Requested band frequency should be between 0.4 and 0.8 microns!')
            bandname = get_bandname_from_wavelength(bandname, rsr)

        wvl, resp = rsr.rsr[bandname][
            'det-1']['wavelength'], rsr.rsr[bandname]['det-1']['response']

        return get_central_wave(wvl, resp, weight=1. / wvl**4)

    def get_poly_coeff(self):
        """Extract the polynomial fit coefficients from file"""

        with h5py.File(self.coeff_filename, 'r') as h5f:
            tab = h5f['coeff'][:]
            wvl = h5f['wavelengths'][:]
            azidiff = h5f['azimuth_difference'][:]

        return tab, wvl, azidiff

    def get_reflectance(self, sun_zenith, sat_zenith, azidiff, bandname,
                        blueband=None):
        """Get the reflectance from the thre sun-sat angles."""
        # Get wavelength in nm for band:
        wvl = self.get_effective_wavelength(bandname) * 1000.0
        coeff, wvl_coord, azid_coord = self.get_poly_coeff()

        if wvl > wvl_coord.max() or wvl < wvl_coord.min():
            LOG.warning(
                "Effective wavelength for band %s outside 400-800 nm range!", str(bandname))
            LOG.info(
                "Set the rayleigh/aerosol reflectance contribution to zero!")
            return np.zeros(sun_zenith.shape)

        idx = np.sqrt((wvl_coord - wvl)**2).argsort()[0]
        wvl1 = wvl_coord[idx]
        wvl2 = wvl_coord[idx + 1]

        shape = sun_zenith.shape
        c1_ = coeff[idx, np.round(azidiff).astype('i').ravel(), :].transpose().reshape(
            21, shape[0], shape[1])
        c2_ = coeff[idx + 1, np.round(azidiff).astype('i').ravel(), :].transpose().reshape(
            21, shape[0], shape[1])
        sun_zenith = np.clip(sun_zenith, 0, 88.)
        res1 = get_rayleigh_reflectance(c1_, sun_zenith, sat_zenith)
        res2 = get_rayleigh_reflectance(c2_, sun_zenith, sat_zenith)

        # Bilinear interpolation
        res = ((res2 - res1) * wvl + res1 * wvl2 - res2 * wvl1) / (wvl2 - wvl1)
        res *= 100

        if blueband is not None:
            res = np.where(np.less(blueband, 20.), res,
                           (1 - (blueband - 20) / 80) * res)

        return np.clip(res, 0, 100)


def get_bandname_from_wavelength(wavelength, rsr):
    """Get the bandname from h5 rsr provided the approximate wavelength."""
    epsilon = 0.1
    # channel_list = [channel for channel in rsr.rsr if abs(
    # rsr.rsr[channel]['det-1']['central_wavelength'] - wavelength) < epsilon]

    chdist_min = 2.0
    chfound = None
    for channel in rsr.rsr:
        chdist = abs(
            rsr.rsr[channel]['det-1']['central_wavelength'] - wavelength)
        if chdist < chdist_min and chdist < epsilon:
            chdist_min = chdist
            chfound = channel

    return BANDNAMES.get(chfound, chfound)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c20671.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Base class for reading raw instrument spectral responses 
"""

import ConfigParser
import os
import numpy as np

import logging
LOG = logging.getLogger(__name__)

try:
    CONFIG_FILE = os.environ['PSP_CONFIG_FILE']
except KeyError:
    LOG.exception('Environment variable PSP_CONFIG_FILE not set!')
    raise

if not os.path.exists(CONFIG_FILE) or not os.path.isfile(CONFIG_FILE):
    raise IOError(str(CONFIG_FILE) + " pointed to by the environment " +
                  "variable PSP_CONFIG_FILE is not a file or does not exist!")


class InstrumentRSR(object):

    """Base class for the raw (agency dependent) instrument response functions"""

    def __init__(self, bandname, platform_name, bandnames=None):
        self.platform_name = platform_name
        self.bandname = bandname
        self.instrument = None
        self.filenames = {}
        self.requested_band_filename = None
        self.bandnames = bandnames
        if not self.bandnames:
            self.bandnames = []
        for band in self.bandnames:
            self.filenames[band] = None
        self.rsr = None

        self.output_dir = None
        self.path = None
        self.options = {}
        self.requested_band_filename = None

    def _get_options_from_config(self):

        conf = ConfigParser.ConfigParser()
        try:
            conf.read(CONFIG_FILE)
        except ConfigParser.NoSectionError:
            LOG.exception('Failed reading configuration file: %s',
                          str(CONFIG_FILE))
            raise

        options = {}
        for option, value in conf.items(self.platform_name + '-' +
                                        self.instrument,
                                        raw=True):
            options[option] = value

        for option, value in conf.items('general', raw=True):
            options[option] = value

        self.output_dir = options.get('rsr_dir', './')
        self.path = options['path']
        self.options = options

    def _get_bandfilenames(self):
        """Get the instrument rsr filenames"""
        for band in self.bandnames:
            LOG.debug("Band = %s", str(band))
            self.filenames[band] = os.path.join(self.path, self.options[band])
            LOG.debug(self.filenames[band])
            if not os.path.exists(self.filenames[band]):
                LOG.warning("Couldn't find an existing file for this band: %s",
                            str(self.filenames[band]))

    def _load(self, scale=1.0):
        """Load the instrument RSR from file(s)"""

        raise NotImplementedError(
            "Instrument rsr loader needs to be defined in the subclass")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2013, 2014, 2015, 2016, 2017 Adam Dybbroe

# Author(s):

#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


try:
    with open('./README', 'r') as fd:
        long_description = fd.read()
except IOError:
    long_description = ''


from setuptools import setup
import imp
import os.path

version = imp.load_source('pyspectral.version', 'pyspectral/version.py')

setup(name='pyspectral',
      version=version.__version__,
      description='Getting satellite sensor rsr functions and the solar spectrum',
      author='Adam Dybbroe',
      author_email='adam.dybbroe@smhi.se',
      classifiers=['Development Status :: 4 - Beta',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: GNU General Public License v3 ' +
                   'or later (GPLv3+)',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python',
                   'Topic :: Scientific/Engineering'],
      url='https://github.com/adybbroe/pyspectral',
      # download_url="https://github.com/adybbroe/py....
      long_description=long_description,
      license='GPLv3',

      packages=['pyspectral'],

      include_package_data=True,
      package_data={
          # If any package contains *.txt files, include them:
          '': ['*.txt', '*.det'],
          'pyspectral': [os.path.join('etc', 'pyspectral.cfg'),
                         os.path.join('data', '*.dat'),
                         os.path.join('data', '*.XLS'),
                         'data/modis/terra/Reference_RSR_Dataset/*.det'],

      },

      # Project should use reStructuredText, so ensure that the docutils get
      # installed or upgraded on the target machine
      install_requires=['docutils>=0.3',
                        'numpy>=1.5.1', 'scipy>=0.8.1',
                        'h5py>=2.5',
                        'requests', 'tqdm'],
      #                  'requests>=2.7.0', 'tqdm>=4.8.4'],
      extras_require={'xlrd': ['xlrd'],
                      'matplotlib': ['matplotlib'],
                      'pandas': ['pandas']},
      scripts=['plot_rsr.py'],
      data_files=[('share', ['pyspectral/data/e490_00a.dat',
                             'pyspectral/data/MSG_SEVIRI_Spectral_Response_Characterisation.XLS'])],
      test_suite='pyspectral.tests.suite',
      tests_require=['mock', 'xlrd'],
      zip_safe=False
      )
